/*
 * Author: Andrew Villadsen
 *
 * Copyright (c) 2017 Trustwave Holdings, Inc.
 * All Rights Reserved.
 *
 * This software is the confidential and proprietary information
 * of Trustwave Holdings, Inc.  Use of this software is governed by
 * the terms and conditions of the license statement and limited
 * warranty furnished with the software.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD TRUSTWAVE HOLDINGS INC.,
 * ITS RELATED COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST
 * ANY CLAIMS OR LIABILITIES ARISING OUT OF OR RESULTING FROM THE USE,
 * MODIFICATION, OR DISTRIBUTION OF PROGRAMS OR FILES CREATED FROM,
 * BASED ON, AND/OR DERIVED FROM THIS SOURCE CODE FILE.
 */

import 'dart:async';

import 'package:angular2/angular2.dart';

import 'package:tw_portal/portal.dart';

import 'package:tw_saved_search_angular/src/common/model.dart';

// TODO: inject Properties, Serializer, and Logger once they are made into Angular services

typedef SavedSearchGroup SavedSearchGroupFactory(String groupName);

SavedSearchGroup createSavedSearchGroup(String groupName) {
  return new _SavedSearchGroupImpl(groupName, Properties.global, Serializer.global, new Logger('SavedSearchGroup'));
}

const SavedSearchGroupFactoryProvider = const Provider(SavedSearchGroupFactory, useValue: createSavedSearchGroup);

abstract class SavedSearchGroup {
  String get name;

  Stream<Iterable<SavedSearch>> get searches;

  Future<SavedSearch> getSearch(String name);

  Future saveSearch(SavedSearch search);

  Future deleteSearch(String searchName);

  factory SavedSearchGroup(String name, Properties properties, Serializer serializer, Logger logger) =
      _SavedSearchGroupImpl;
}

class _SavedSearchGroupImpl implements SavedSearchGroup {
  final Properties _properties;
  final Serializer _serializer;
  final Logger _log;

  Iterable<SavedSearch> _currentSearches;

  StreamController<Iterable<SavedSearch>> _searches;
  Stream<Iterable<SavedSearch>> get searches {
    if (_searches == null) {
      _searches = new StreamController(onListen: _loadSearches);
    }
    return _searches.stream;
  }

  final String name;

  _SavedSearchGroupImpl(this.name, Properties properties, this._serializer, this._log)
      : _properties = properties.withPrefix('portal-ng.public.saved-search') {
    SavedSearch.registerJson(_serializer);
  }

  Future<SavedSearch> getSearch(String name) async {
    await _loadSearches();
    return _currentSearches.firstWhere((search) => search.searchName == name, orElse: () => null);
  }

  Future saveSearch(SavedSearch search) async {
    await _loadSearches();

    // Take a copy since the _currentSearches reference is exposed and we don't want to change it
    List<SavedSearch> newSearches = _currentSearches.toList();

    bool alreadyExists = false;
    for (var index = 0; index < newSearches.length; index++) {
      if (newSearches[index].searchName == search.searchName) {
        newSearches[index] = search;
        alreadyExists = true;
        break;
      }
    }
    if (!alreadyExists) {
      newSearches.add(search);
    }

    return _save(newSearches);
  }

  Future deleteSearch(String searchName) async {
    await _loadSearches();

    return _save(_currentSearches.where((search) => search.searchName != searchName).toList());
  }

  bool _loaded = false;
  Future _loadSearches() async {
    // Only perform the initial load once
    if (_loaded) {
      return;
    }
    _loaded = true;

    await _properties.load(name);

    final String encodedSearches = _properties.get(name, "");
    _updateSearches(encodedSearches.isEmpty ? [] : (_serializer.decodeRpc(encodedSearches) ?? []));
  }

  Future _save(List<SavedSearch> newSearches) async {
    Iterable<SavedSearch> oldSearches = _currentSearches;
    _updateSearches(newSearches);
    try {
      await _properties.save(name, _serializer.encodeRpc(_currentSearches));
    } catch (error) {
      _log.error('Failed to save searches with group name "$name"', error);
      _updateSearches(oldSearches);
      rethrow;
    }
  }

  void _updateSearches(Iterable<SavedSearch> newSearches) {
    _currentSearches = newSearches;
    _searches?.add(_currentSearches);
  }
}
