/*
 * Author: Andrew Villadsen
 *
 * Copyright (c) 2017 Trustwave Holdings, Inc.
 * All Rights Reserved.
 *
 * This software is the confidential and proprietary information
 * of Trustwave Holdings, Inc.  Use of this software is governed by
 * the terms and conditions of the license statement and limited
 * warranty furnished with the software.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD TRUSTWAVE HOLDINGS INC.,
 * ITS RELATED COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST
 * ANY CLAIMS OR LIABILITIES ARISING OUT OF OR RESULTING FROM THE USE,
 * MODIFICATION, OR DISTRIBUTION OF PROGRAMS OR FILES CREATED FROM,
 * BASED ON, AND/OR DERIVED FROM THIS SOURCE CODE FILE.
 */

import 'dart:async';

import 'package:angular2/angular2.dart';

import 'package:tw_portal/portal.dart';
import 'package:tw_ui_angular/tw_ui_angular.dart';

import 'package:tw_saved_search_angular/src/common/model.dart';
import 'package:tw_saved_search_angular/src/common/service.dart';
import 'package:tw_saved_search_angular/src/ui/saved_search/saved_search_event.dart';

@Component(selector: 'twa-saved-search', templateUrl: 'saved_search.html', directives: const [
  NgFor,
  MenuComponent,
  MenuGroupComponent,
  MenuItemComponent,
  DialogComponent,
  AlertComponent,
  FormComponent,
  FormItemComponent,
  InputComponent,
], providers: const [
  SavedSearchGroupFactoryProvider
])
class SavedSearchComponent {
  //
  // Inputs
  //

  @Input()
  set groupName(String name) {
    if (groupName != name) {
      _searchGroup = _searchGroupFactory(name);
    }
  }

  String get groupName => _searchGroup?.name;

  @Input()
  set search(value) {
    if (_search != value) {
      _search = value;
      _searchChange.add(value);
    }
  }

  dynamic get search => _search;
  dynamic _search;

  @Input()
  String label;

  //
  // Outputs
  //

  @Output()
  Stream<dynamic> get searchChange => _searchChange.stream;
  final StreamController<dynamic> _searchChange = new StreamController<dynamic>.broadcast();

  @Output()
  Stream<SavedSearchEvent> get syntaxTips => _syntaxTips.stream;
  final StreamController<SavedSearchEvent> _syntaxTips = new StreamController<SavedSearchEvent>.broadcast(sync: true);

  //
  // Services & Constructor
  //

  final SavedSearchGroupFactory _searchGroupFactory;

  SavedSearchComponent(this._searchGroupFactory);

  //
  // Template Members
  //

  bool get isPromptingNew => _prompt == _Prompt.NEW;
  bool get isPromptingChanges => _prompt == _Prompt.CHANGES;
  bool get isPromptingDelete => _prompt == _Prompt.DELETE;

  @ViewChild(InputComponent)
  InputComponent searchNameInput;

  String selectedSearchName;
  String inputSearchName;

  Stream<Iterable<SavedSearch>> get savedSearches => _searchGroup?.searches;

  // Track by function for searches list
  // first parameter is unused index
  String savedSearchName(_, SavedSearch savedSearch) => savedSearch.searchName;

  Future save() async {
    if (isPromptingNew) {
      searchNameInput.validate();
      if (searchNameInput.valid) {
        // Check for duplicate search name.
        // Can't use the validation function since this is asynchronous.
        if ((await _searchGroup.getSearch(inputSearchName)) != null) {
          searchNameInput.validationMsg = duplicateSearchNameMsg;
          searchNameInput.valid = false;
        }
      }

      if (!searchNameInput.valid) {
        return;
      }
    }

    final savedSearch = new SavedSearch();
    savedSearch.searchName = inputSearchName;
    savedSearch.searchCriteria = search;

    await _searchGroup.saveSearch(savedSearch);
    selectedSearchName = inputSearchName;
    inputSearchName = '';
    _prompt = _Prompt.NONE;
  }

  void cancelSave() {
    inputSearchName = '';
    _prompt = _Prompt.NONE;
  }

  Future delete() async {
    await _searchGroup.deleteSearch(selectedSearchName);
    selectedSearchName = null;
    _prompt = _Prompt.NONE;
  }

  void cancelDelete() {
    _prompt = _Prompt.NONE;
  }

  void select(SavedSearch savedSearch) {
    selectedSearchName = savedSearch.searchName;
    search = savedSearch.searchCriteria;
  }

  void promptNew() {
    selectedSearchName = null;
    _prompt = _Prompt.NEW;
  }

  void promptChanges() {
    inputSearchName = selectedSearchName;
    _prompt = _Prompt.CHANGES;
  }

  void promptDelete() {
    _prompt = _Prompt.DELETE;
  }

  void requestSyntaxTips() {
    _syntaxTips.add(new SavedSearchEvent());
  }

  // TODO: localize
  String get menuLabel => label ?? 'Select...';
  String get actionsGroupLabel => 'Save a Search';
  String get saveAsNewLabel => 'Save As New Search...';
  String get saveChangesLabel => 'Save Changes';
  String get deleteLabel => 'Delete';
  String get syntaxTipsLabel => 'Syntax Tips...';
  String get saveAsNewHeading => 'Save Search';
  String get saveChangesHeading => 'Save Changes';
  String get searchNameLabel => 'Search Name';
  String get deleteHeading => 'Delete A Saved Search';
  String get deleteMessage => "Are you sure you want to delete the saved search '$selectedSearchName'?";
  String get duplicateSearchNameMsg => 'search name already exists';

  //
  // Private Members
  //

  SavedSearchGroup _searchGroup;

  _Prompt _prompt;
}

enum _Prompt { NONE, NEW, CHANGES, DELETE }
