/*
 * Author: Andrew Villadsen
 *
 * Copyright (c) 2017 Trustwave Holdings, Inc.
 * All Rights Reserved.
 *
 * This software is the confidential and proprietary information
 * of Trustwave Holdings, Inc.  Use of this software is governed by
 * the terms and conditions of the license statement and limited
 * warranty furnished with the software.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD TRUSTWAVE HOLDINGS INC.,
 * ITS RELATED COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST
 * ANY CLAIMS OR LIABILITIES ARISING OUT OF OR RESULTING FROM THE USE,
 * MODIFICATION, OR DISTRIBUTION OF PROGRAMS OR FILES CREATED FROM,
 * BASED ON, AND/OR DERIVED FROM THIS SOURCE CODE FILE.
 */

import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import 'package:tw_portal/portal.dart';

import 'package:tw_saved_search_angular/src/common/model.dart';
import 'package:tw_saved_search_angular/src/common/service.dart';

void main() {
  final Properties properties = new MockProperties();
  final Serializer serializer = new MockSerializer();
  final Logger logger = new MockLogger();

  SavedSearchGroup searchGroup;

  setUp(() {
    when(properties.withPrefix(any)).thenReturn(properties);
    when(properties.get(any, any)).thenReturn('');

    searchGroup = new SavedSearchGroup('foo', properties, serializer, logger);
  });

  tearDown(() {
    reset(properties);
    reset(serializer);
    reset(logger);
  });

  test('name', () {
    expect(searchGroup.name, 'foo');
  });

  test('uses saved search property prefix', () {
    verify(properties.withPrefix('portal-ng.public.saved-search')).called(1);
  });

  group('initial load', () {
    Iterable<SavedSearch> searches;

    setUp(() async {
      final SavedSearch savedSearch = new SavedSearch();
      savedSearch.searchName = 'name';

      when(properties.get(any, any)).thenReturn('bar');
      when(serializer.decodeRpc(any)).thenReturn([savedSearch]);

      // Note: this actually triggers the initial load
      searches = await searchGroup.searches.first;
    });

    test('loads the saved search group', () async {
      verify(properties.load('foo')).called(1);
    });

    test('gets the saved search group', () {
      verify(properties.get('foo', any)).called(1);
    });

    test('deserializes the group property', () {
      verify(serializer.decodeRpc('bar')).called(1);
    });

    test('publishes the retrieved searches', () async {
      expect(searches, hasLength(1));
      expect(searches.first.searchName, 'name');
    });
  });

  group('getSearch', () {
    setUp(() {
      final SavedSearch existingSearch = new SavedSearch();
      existingSearch.searchName = 'existing';
      existingSearch.searchCriteria = 1;

      when(properties.get(any, any)).thenReturn('anything not empty');
      when(serializer.decodeRpc(any)).thenReturn([existingSearch]);
    });

    test('gets existing search', () async {
      SavedSearch existing = await searchGroup.getSearch('existing');

      expect(existing.searchName, 'existing');
      expect(existing.searchCriteria, 1);
    });

    test('completes with null if search does not exist', () async {
      expect(await searchGroup.getSearch('does not exist'), isNull);
    });
  });

  group('saveSearch', () {
    setUp(() {
      final SavedSearch existingSearch = new SavedSearch();
      existingSearch.searchName = 'existing';
      existingSearch.searchCriteria = 1;

      when(properties.get(any, any)).thenReturn('anything not empty');
      when(serializer.decodeRpc(any)).thenReturn([existingSearch]);
      when(serializer.encodeRpc(any)).thenReturn('serialized searches');
    });

    test('updates an existing search', () async {
      SavedSearch updated = new SavedSearch();
      updated.searchName = 'existing';
      updated.searchCriteria = 2;

      await searchGroup.saveSearch(updated);

      expect(
          (verify(serializer.encodeRpc(captureAny)).captured.first as Iterable<SavedSearch>),
          allOf(
              hasLength(1),
              contains(
                  predicate((SavedSearch search) => search.searchName == 'existing' && search.searchCriteria == 2))));
      verify(properties.save('foo', 'serialized searches')).called(1);
    });

    test('adds a new search', () async {
      SavedSearch newSearch = new SavedSearch();
      newSearch.searchName = 'new';
      newSearch.searchCriteria = 2;

      await searchGroup.saveSearch(newSearch);

      expect(
          (verify(serializer.encodeRpc(captureAny)).captured.first as Iterable<SavedSearch>),
          allOf(
              hasLength(2),
              containsAllInOrder([
                predicate((SavedSearch search) => search.searchName == 'existing' && search.searchCriteria == 1),
                predicate((SavedSearch search) => search.searchName == 'new' && search.searchCriteria == 2)
              ])));
    });

    test('publishes searches', () async {
      SavedSearch updated = new SavedSearch();
      updated.searchName = 'existing';
      updated.searchCriteria = 2;

      expect(
          searchGroup.searches,
          emitsThrough(allOf(
              hasLength(1),
              contains(
                  predicate((SavedSearch search) => search.searchName == 'existing' && search.searchCriteria == 2)))));

      await searchGroup.saveSearch(updated);
    });

    test('publishes old searches again on error', () async {
      SavedSearch updated = new SavedSearch();
      updated.searchName = 'existing';
      updated.searchCriteria = 2;
      when(properties.save(any, any)).thenThrow(new Error());

      expect(
          searchGroup.searches,
          emitsInOrder([
            allOf(
                hasLength(1),
                contains(
                    predicate((SavedSearch search) => search.searchName == 'existing' && search.searchCriteria == 1))),
            allOf(
                hasLength(1),
                contains(
                    predicate((SavedSearch search) => search.searchName == 'existing' && search.searchCriteria == 2))),
            allOf(
                hasLength(1),
                contains(
                    predicate((SavedSearch search) => search.searchName == 'existing' && search.searchCriteria == 1)))
          ]));

      expect(searchGroup.saveSearch(updated), throwsA(anything));
    });
  });

  group('deleteSearch', () {
    setUp(() {
      final SavedSearch searchToDelete = new SavedSearch();
      searchToDelete.searchName = 'deleted';
      final SavedSearch searchNotToDelete = new SavedSearch();
      searchNotToDelete.searchName = 'not deleted';
      when(properties.get(any, any)).thenReturn('anything not empty');
      when(serializer.decodeRpc(any)).thenReturn([searchToDelete, searchNotToDelete]);
      when(serializer.encodeRpc(any)).thenReturn('serialized searches');
    });

    test('removes the search with given name', () async {
      await searchGroup.deleteSearch('deleted');

      expect((verify(serializer.encodeRpc(captureAny)).captured.first as Iterable<SavedSearch>).single.searchName,
          'not deleted');
      verify(properties.save('foo', 'serialized searches')).called(1);
    });

    test('saves the saved search group property', () async {
      await searchGroup.deleteSearch('deleted');

      verify(properties.save('foo', 'serialized searches')).called(1);
    });

    test('publishes searches', () async {
      expect(
          searchGroup.searches,
          emitsThrough(
              allOf([hasLength(1), contains(predicate((SavedSearch search) => search.searchName == 'not deleted'))])));

      await searchGroup.deleteSearch('deleted');
    });

    test('publishes old searches again on error', () async {
      when(properties.save(any, any)).thenThrow(new Error());

      expect(searchGroup.searches, emitsInOrder([hasLength(2), hasLength(1), hasLength(2)]));

      expect(searchGroup.deleteSearch('deleted'), throwsA(anything));
    });
  });
}

class MockProperties extends Mock implements Properties {}

class MockSerializer extends Mock implements Serializer {}

class MockLogger extends Mock implements Logger {}
