/*
 * Author: Andrew Villadsen
 *
 * Copyright (c) 2017 Trustwave Holdings, Inc.
 * All Rights Reserved.
 *
 * This software is the confidential and proprietary information
 * of Trustwave Holdings, Inc.  Use of this software is governed by
 * the terms and conditions of the license statement and limited
 * warranty furnished with the software.
 *
 * IN PARTICULAR, YOU WILL INDEMNIFY AND HOLD TRUSTWAVE HOLDINGS INC.,
 * ITS RELATED COMPANIES AND ITS SUPPLIERS, HARMLESS FROM AND AGAINST
 * ANY CLAIMS OR LIABILITIES ARISING OUT OF OR RESULTING FROM THE USE,
 * MODIFICATION, OR DISTRIBUTION OF PROGRAMS OR FILES CREATED FROM,
 * BASED ON, AND/OR DERIVED FROM THIS SOURCE CODE FILE.
 */

@Tags(const ['aot'])

import 'dart:async';

import 'package:angular2/angular2.dart';
import 'package:angular_components/angular_components.dart';
import 'package:angular_test/angular_test.dart';
import 'package:mockito/mirrors.dart';
import 'package:test/test.dart';

import 'package:tw_ui_angular/tw_ui_angular.dart';

import 'package:tw_saved_search_angular/src/common/model.dart';
import 'package:tw_saved_search_angular/src/common/service.dart';
import 'package:tw_saved_search_angular/src/ui/saved_search/saved_search.dart';

import 'saved_search_po.dart';

@AngularEntrypoint()
void main() {
  final SavedSearchGroup searchGroup = new MockSavedSearchGroup();
  String factoryCalledName;
  SavedSearchGroup mockSearchGroupFactory(String name) {
    factoryCalledName = name;
    return searchGroup;
  }

  StreamController<Iterable<SavedSearch>> searches;

  final NgTestBed<TestComponent> testBed = new NgTestBed<TestComponent>()
      .addProviders([materialProviders, new Provider(SavedSearchGroupFactory, useValue: mockSearchGroupFactory)]);
  NgTestFixture<TestComponent> fixture;
  SavedSearchPO savedSearchPO;
  SavedSearchMenuPO menuPO;

  setUp(() {
    searches = new StreamController<Iterable<SavedSearch>>();
    when(searchGroup.searches).thenReturn(searches.stream);

    final SavedSearch search1 = new SavedSearch();
    search1.searchName = 'existing 1';
    search1.searchCriteria = 1;
    final SavedSearch search2 = new SavedSearch();
    search2.searchName = 'existing 2';
    search2.searchCriteria = 2;

    searches.add([search1, search2]);
  });

  setUp(() async {
    fixture = await testBed.create(beforeChangeDetection: (search) => search.groupName = 'foo');

    savedSearchPO = await fixture.resolvePageObject(SavedSearchPO);
    await savedSearchPO.open();
    menuPO = await savedSearchPO.resolveMenu(fixture);
  });

  tearDown(() {
    reset(searchGroup);
    factoryCalledName = null;
  });

  tearDown(disposeAnyRunningTest);

  test('setting groupName creates new search group', () async {
    expect(factoryCalledName, 'foo');
  });

  test('label', () async {
    await fixture.update((search) => search.label = 'foo');
    savedSearchPO = await fixture.resolvePageObject(SavedSearchPO);

    expect(await savedSearchPO.label, 'foo');
  });

  test('displays searches', () async {
    expect(await Future.wait(menuPO.searches.map((search) => search.name)), ['existing 1', 'existing 2']);
  });

  test('displays most recent searches', () async {
    SavedSearch search = new SavedSearch();
    search.searchName = 'new';

    searches.add([search]);
    menuPO = await savedSearchPO.resolveMenu(fixture);

    expect(menuPO.searches, hasLength(1));
    expect(await menuPO.searches.first.name, 'new');
  });

  test('no search is selected at first', () async {
    expect(menuPO.searches.where((search) => search.selected), hasLength(0));
  });

  group('selecting a search', () {
    test('shows that search as selected', () async {
      // I'm not sure why, but this test doesn't stabilize unless we wait for a fixture.update() here
      await fixture.update();
      await menuPO.searches.first.select();

      menuPO = await savedSearchPO.resolveMenu(fixture);
      expect(menuPO.searches.first.selected, isTrue);
    });

    test('publishes searchChange event', () async {
      await fixture.update((search) => expect(search.searchChange, emits(1)));

      await menuPO.searches.first.select();
    });
  });

  group('save as new', () {
    SaveDialogPO saveDialogPO;

    setUp(() async {
      await fixture.update((comp) => comp.search = 3);

      await menuPO.saveAsNew();

      saveDialogPO = await savedSearchPO.resolveSaveDialog(fixture);
    });

    test('adds new saved search', () async {
      await saveDialogPO.typeName('new search');
      await saveDialogPO.save();

      SavedSearch saved = verify(searchGroup.saveSearch(captureAny)).captured.single;
      expect(saved.searchName, 'new search');
      expect(saved.searchCriteria, 3);
    });

    test('cancel does not add search', () async {
      await saveDialogPO.cancel();

      verifyNever(searchGroup.saveSearch(any));
    });

    test('does not accept empty name', () async {
      await saveDialogPO.clearName();
      await saveDialogPO.save();

      verifyNever(searchGroup.saveSearch(any));
    });

    test('does not accept existing name', () async {
      when(searchGroup.getSearch('existing 1')).thenReturn(new SavedSearch()..searchName = 'existing 1');

      await saveDialogPO.typeName('existing 1');
      await saveDialogPO.save();

      verifyNever(searchGroup.saveSearch(any));
    });
  });

  group('save changes', () {
    SaveDialogPO saveDialogPO;

    setUp(() async {
      // I'm not sure why, but these tests don't stabilize unless we wait for a fixture.update() here
      await fixture.update();
    });

    test('modifies selected search', () async {
      await menuPO.searches.first.select();
      await fixture.update((comp) => comp.search = 3);
      await savedSearchPO.open();
      await menuPO.saveChanges();
      saveDialogPO = await savedSearchPO.resolveSaveDialog(fixture);

      saveDialogPO.save();

      // For whatever reason, this test seems to need a moment to stabilize
      await fixture.update();

      SavedSearch saved = verify(searchGroup.saveSearch(captureAny)).captured.single;
      expect(saved.searchName, 'existing 1');
      expect(saved.searchCriteria, 3);
    }, timeout: new Timeout(new Duration(seconds: 30)));

    test('is disabled if no search is selected', () async {
      expect(await menuPO.saveChangesDisabled, isTrue);
    });

    test('cancel does not modify search', () async {
      await menuPO.searches.first.select();
      await savedSearchPO.open();
      await menuPO.saveChanges();
      saveDialogPO = await savedSearchPO.resolveSaveDialog(fixture);

      saveDialogPO.cancel();

      verifyNever(searchGroup.saveSearch(any));
    });
  });

  group('delete', () {
    DeleteAlertPO deleteAlertPO;

    test('calls delete on selected search', () async {
      // I'm not sure why, but this test doesn't stabilize unless we wait for a fixture.update() here
      await fixture.update();

      await menuPO.searches.first.select();
      await savedSearchPO.open();
      await menuPO.delete();
      deleteAlertPO = await savedSearchPO.resolveDeleteAlert(fixture);
      await deleteAlertPO.delete();

      verify(searchGroup.deleteSearch('existing 1')).called(1);
    });

    test('is disabled if no search is selected', () async {
      expect(await menuPO.deleteDisabled, isTrue);
    });

    test('cancel does not delete search', () async {
      // I'm not sure why, but this test doesn't stabilize unless we wait for a fixture.update() here
      await fixture.update();

      await menuPO.searches.first.select();
      await savedSearchPO.open();
      await menuPO.delete();
      deleteAlertPO = await savedSearchPO.resolveDeleteAlert(fixture);
      await deleteAlertPO.cancel();

      verifyNever(searchGroup.deleteSearch(any));
    });
  });

  test('syntax tips', () async {
    savedSearchPO = await fixture.resolvePageObject(SavedSearchPO);

    await fixture.update((search) => expect(search.syntaxTips, emits(anything)));

    await savedSearchPO.open();
    menuPO = await savedSearchPO.resolveMenu(fixture);
    await menuPO.syntaxTips();
  });
}

class MockSavedSearchGroup extends Mock implements SavedSearchGroup {}

@Component(
  selector: 'twa-saved-search',
  templateUrl: 'package:tw_saved_search_angular/src/ui/saved_search/saved_search.html',
  styleUrls: const ['package:tw_saved_search_angular/src/ui/saved_search/saved_search.css'],
  directives: const [twaDirectives],
  inputs: const [
    'groupName',
    'search',
    'label',
  ],
  outputs: const [
    'searchChange',
    'syntaxTips',
  ],
)
class TestComponent extends SavedSearchComponent {
  TestComponent(SavedSearchGroupFactory searchGroupFactory) : super(searchGroupFactory);

  // Using Component.queries doesn't work for some reason
  @ViewChild(InputComponent)
  set searchNameInput(InputComponent input) => super.searchNameInput = input;
}
